from __future__ import print_function
import warnings
warnings.filterwarnings('ignore')

import math
import torch
from dataset_prep import Dataset_MMD
from torch.utils.data import DataLoader
from sklearn.metrics import confusion_matrix
from torch import nn
from torch import optim
from net.st_gcn import Model_stu, Model_tea1, Model_tea2
import random
import torch.nn.functional as F
import numpy as np
from sklearn.metrics import roc_auc_score


def step_decay(epoch, learning_rate, drop, epochs_drop):
    """
    learning rate step decay
    :param epoch: current training epoch
    :param learning_rate: initial learning rate
    :return: learning rate after step decay
    """
    initial_lrate = learning_rate
    lrate = initial_lrate * math.pow(drop, math.floor((1 + epoch) / epochs_drop))
    return lrate


def train_ddcnet_union(epoch, ddcnet_stu, ddcnet_tea1, ddcnet_tea2, src_loader, tgt1_train_loader, tgt2_train_loader, learning_rate, drop, epochs_drop):
    log_interval = 1  # original: 10
    LEARNING_RATE = step_decay(epoch, learning_rate, drop, epochs_drop)
    print(f'Learning Rate: {LEARNING_RATE}')
    optimizer = optim.SGD(ddcnet_stu.parameters(), lr=LEARNING_RATE, momentum=MOMENTUM, weight_decay=L2_DECAY)
    stu_clf_criterion = nn.BCELoss()
    ddcnet_stu.train()

    train_correct = 0
    train_loss = 0
    TN = 0
    FP = 0
    FN = 0
    TP = 0

    tr_stu_src_auc_gt = []
    tr_stu_src_auc_pred = []

    len_dataloader = min(len(src_loader), len(tgt1_train_loader), len(tgt2_train_loader))
    for step, (src_batch, tgt1_batch, tgt2_batch) in enumerate(zip(src_loader, tgt1_train_loader, tgt2_train_loader)):
        optimizer.zero_grad()
        src_data_batch = src_batch['data'].to(device).float()
        src_label_batch = src_batch['label'].to(device).float()
        tgt1_data_batch = tgt1_batch['data'].to(device).float()
        tgt2_data_batch = tgt2_batch['data'].to(device).float()

        stu_src_pred, _ = ddcnet_stu(src_data_batch, 0)
        stu_tgt1_pred, _ = ddcnet_stu(tgt1_data_batch, 0)
        stu_tgt2_pred, _ = ddcnet_stu(tgt2_data_batch, 0)

        with torch.no_grad():
            tea1_src_pred, _ = ddcnet_tea1(src_data_batch, 0)
            tea1_tgt1_pred, _ = ddcnet_tea1(tgt1_data_batch, 0)
            tea2_src_pred, _ = ddcnet_tea2(src_data_batch, 0)
            tea2_tgt2_pred, _ = ddcnet_tea2(tgt2_data_batch, 0)

        # total loss = (1) classification loss + (2) consistency loss.
        # Mean teacher does not need parameter loss, since teacher's parameters is EMA of student's parameters

        # (1) classification loss
        stu_clf_loss = stu_clf_criterion(stu_src_pred[:, 0], src_label_batch)
        clf_loss = stu_clf_loss

        # (2) consistency loss (MSE or KLD)
        consistency_loss_mse = F.mse_loss(stu_src_pred[:, 0], tea1_src_pred[:, 0]) + F.mse_loss(stu_src_pred[:, 0], tea2_src_pred[:, 0]) + F.mse_loss(stu_tgt1_pred[:, 0], tea1_tgt1_pred[:, 0]) + F.mse_loss(stu_tgt2_pred[:, 0], tea2_tgt2_pred[:, 0])
        consistency_loss_kld = F.kl_div((F.softmax(stu_src_pred[:, 0], dim=-1)).log(), F.softmax(tea1_src_pred[:, 0], dim=-1)) + F.kl_div((F.softmax(stu_src_pred[:, 0], dim=-1)).log(), F.softmax(tea2_src_pred[:, 0], dim=-1)) + F.kl_div((F.softmax(stu_tgt1_pred[:, 0], dim=-1)).log(), F.softmax(tea1_tgt1_pred[:, 0], dim=-1)) + F.kl_div((F.softmax(stu_tgt2_pred[:, 0], dim=-1)).log(), F.softmax(tea2_tgt2_pred[:, 0], dim=-1))

        loss = 0.1 * clf_loss + 10 * consistency_loss_kld + 1 * consistency_loss_mse

        loss.backward()
        optimizer.step()
        train_loss += loss

        # update mean teacher
        alpha_tea1 = 0.99
        alpha_tea2 = 0.99
        # alpha_tea1 = min(0.05 + 0.1 * math.floor(epoch/float(2)), 0.9)
        # alpha_tea2 = min(0.05 + 0.1 * math.floor(epoch/float(2)), 0.9)
        for tea1_param, stu_param in zip(ddcnet_tea1.parameters(), ddcnet_stu.parameters()):
            tea1_param.data.mul_(alpha_tea1).add_(1 - alpha_tea1, stu_param.data)  # θ'(t) = αθ'(t-1) + (1-α)θ(t)

        for tea2_param, stu_param in zip(ddcnet_tea2.parameters(), ddcnet_stu.parameters()):
            tea2_param.data.mul_(alpha_tea2).add_(1 - alpha_tea2, stu_param.data)  # θ'(t) = αθ'(t-1) + (1-α)θ(t)

        stu_src_pred_cpu = stu_src_pred.data.cpu().numpy() > 0.511
        train_correct += sum(stu_src_pred_cpu[:, 0] == src_label_batch.cpu().numpy())  # correct num in one batch

        TN_tmp, FP_tmp, FN_tmp, TP_tmp = confusion_matrix(src_label_batch.cpu().numpy(), stu_src_pred_cpu[:, 0], labels=[0, 1]).ravel()
        TN += TN_tmp
        FP += FP_tmp
        FN += FN_tmp
        TP += TP_tmp

        # tr auc
        tr_stu_src_auc_gt.extend(src_label_batch.cpu().numpy())
        tr_stu_src_auc_pred.extend(stu_src_pred.detach().cpu().numpy())

        if (step + 1) % log_interval == 0:
            print("Train Epoch [{:4d}/{}] Step [{:2d}/{}]: clf_loss={:.6f}, consistency_loss_kld={:.6f}, consistency_loss_mse={:.6f}, loss={:.6f}".format(
                    epoch, TRAIN_EPOCHS, step + 1, len_dataloader, clf_loss.data, consistency_loss_kld.data, consistency_loss_mse.data, loss.data))

    train_loss /= len_dataloader
    train_acc = (TP+TN) / (TP + FP + TN + FN)
    train_F1 = (2 * TP) / (2 * TP + FP + TN)
    train_AUC = roc_auc_score(tr_stu_src_auc_gt, tr_stu_src_auc_pred)

    print('Train set: Average classification loss: {:.4f}, Accuracy: {}/{} ({:.2f}%), train_AUC: {:.5}, train_F1: {:.5}'.format(
        train_loss, train_correct, (len_dataloader * BATCH_SIZE), train_acc, train_AUC, train_F1))

    # save checkpoint.pth, save train loss and acc to a txt file
    if epoch == 80:
        torch.save(ddcnet_stu.state_dict(), SAVE_PATH + 'fold_' + str(fold) + '_epoch_' + str(epoch) + '.pth')
    with open(SAVE_PATH + 'fold_' + str(fold) + '_train_loss_and_acc.txt', 'a') as f:
        f.write('epoch {}, loss {:.5}, train acc {:.5}, train_AUC {:.5}, train_F1: {:.5}\n'.format(epoch, train_loss, train_acc, train_AUC, train_F1))


def test_ddcnet(ddcnet_stu, target_test_loader) -> object:
    with torch.no_grad():
        clf_criterion = nn.BCELoss()  # not nn.CrossEntropyLoss()

        ddcnet_stu.eval()
        test_loss = 0
        test_correct = 0
        TN = 0
        FP = 0
        FN = 0
        TP = 0

        te_stu_auc_gt = []
        te_stu_auc_pred = []

        for step, test_sample_batch in enumerate(target_test_loader):
            test_data_batch = test_sample_batch['data'].to(device).float()
            test_label_batch = test_sample_batch['label'].to(device).float()

            # three outputs should be the same
            stu_test_pred, _ = ddcnet_stu(test_data_batch, 0)

            test_loss += clf_criterion(stu_test_pred[:, 0], test_label_batch)

            test_pred_cpu = stu_test_pred.data.cpu().numpy() > 0.511
            test_correct += sum(test_pred_cpu[:, 0] == test_label_batch.cpu().numpy())
            TN_tmp, FP_tmp, FN_tmp, TP_tmp = confusion_matrix(test_label_batch.cpu().numpy(), test_pred_cpu[:, 0], labels=[0, 1]).ravel()
            TN += TN_tmp
            FP += FP_tmp
            FN += FN_tmp
            TP += TP_tmp

            te_stu_auc_gt.extend(test_label_batch.cpu().numpy())
            te_stu_auc_pred.extend(stu_test_pred.detach().cpu().numpy())

        test_loss /= len(target_test_loader)

        TPR = TP / (TP + FN)  # Sensitivity/ hit rate/ recall/ true positive rate
        TNR = TN / (TN + FP)  # Specificity/ true negative rate
        PPV = TP / (TP + FP)  # Precision/ positive predictive value
        NPV = TN / (TN + FN)  # Negative predictive value
        FPR = FP / (FP + TN)  # Fall out/ false positive rate
        FNR = FN / (TP + FN)  # False negative rate
        FDR = FP / (TP + FP)  # False discovery rate
        test_acc = (TP+TN) / (TP + FP + TN + FN)  # accuracy of each class
        # test_acc = float(test_correct) * 100. / len(target_test_loader)  # another way to compute 'test_acc'
        test_F1 = (2 * TP) / (2 * TP + FP + TN)
        test_AUC = roc_auc_score(te_stu_auc_gt, te_stu_auc_pred)

        print('Test set: Correct_num: {}, test_acc: {:.4f}, test_AUC: {:.4f}, test_F1: {:.4f}, TPR: {:.4f}, TNR: {:.4f}, PPV:{:.4f}, NPV: {:.4f}, FPR: {:.4f}, FNR: {:.4f}, FDR: {:.4f}'.format(
            test_correct, test_acc, test_AUC, test_F1, TPR, TNR, PPV, NPV, FPR, FNR, FDR))

        # save test loss and acc to a txt file
        with open(SAVE_PATH + flag + '_fold_' + str(fold) + '_test_loss_and_acc.txt', 'a') as f:
            f.write('epoch {}, test_acc {:.5}, test_AUC {:.5}, test_F1: {:.4f}, TPR {:.5}, TNR {:.5}, PPV {:.5}, NPV {:.5}, FPR {:.5}, FNR {:.5}, FDR {:.5}\n'.format(epoch, test_acc, test_AUC, test_F1, TPR, TNR, PPV, NPV, FPR, FNR, FDR))


if __name__ == '__main__':
    # m1_preprocess_raw
    # m2_preprocess_norm_allROI_eachsubj
    # m3_preprocess_norm_eachROI_eachsubj
    # m4_preprocess_norm_eachROI_allsubj

    ROOT_PATH = '/home/yuqifang/projects/MICCAI2022_summary/multarget_data/npy_src_20_tgt1_21_tgt2_1/'
    SAVE_PATH = '/home/yuqifang/projects/MICCAI2022_summary/multarget_code/MFH/s2_joint_train_tea1_tea2_stu/checkpoints/'

    BATCH_SIZE = 64
    TRAIN_EPOCHS = 100
    L2_DECAY = 5e-4
    MOMENTUM = 0.9
    learning_rate = 0.01
    drop = 0.5
    epochs_drop = 30.0

    for fold in [3, 8, 9, 13, 30]:
        seed = fold  # fold
        np.random.seed(seed)
        random.seed(seed)
        torch.manual_seed(seed)  # cpu
        torch.cuda.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)  # if you are using multi-GPU.
        torch.backends.cudnn.benchmark = False
        torch.backends.cudnn.deterministic = True  # cpu/gpu结果一致

        print('Load data begin:')
        src_data = ROOT_PATH + 'src_data_aal.npy'  # (srcsubj_num, 1, T, node_num, 1)
        src_label = ROOT_PATH + 'src_label.npy'  # (srcsubj_num,)
        tgt1_data = ROOT_PATH + 'tgt1_data_aal.npy'  # (tgt1subj_num, 1, T, node_num, 1)
        tgt1_label = ROOT_PATH + 'tgt1_label.npy'  # (tgt1subj_num,)
        tgt2_data = ROOT_PATH + 'tgt2_data_aal.npy'  # (tgt2subj_num, 1, T, node_num, 1)
        tgt2_label = ROOT_PATH + 'tgt2_label.npy'  # (tgt2subj_num,)

        src_dataset = Dataset_MMD(src_data, src_label, transform=None)
        src_loader = DataLoader(src_dataset, batch_size=BATCH_SIZE, shuffle=True, drop_last=True)
        src_test_loader = DataLoader(src_dataset, batch_size=BATCH_SIZE, shuffle=False, drop_last=False)

        tgt1_dataset = Dataset_MMD(tgt1_data, tgt1_label, transform=None)
        tgt1_train_loader = DataLoader(tgt1_dataset, batch_size=BATCH_SIZE, shuffle=True, drop_last=True)
        tgt1_test_loader = DataLoader(tgt1_dataset, batch_size=BATCH_SIZE, shuffle=False, drop_last=False)

        tgt2_dataset = Dataset_MMD(tgt2_data, tgt2_label, transform=None)
        tgt2_train_loader = DataLoader(tgt2_dataset, batch_size=BATCH_SIZE, shuffle=True, drop_last=True)
        tgt2_test_loader = DataLoader(tgt2_dataset, batch_size=BATCH_SIZE, shuffle=False, drop_last=False)

        device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
        print('device:', device)

        pretrain_dir = '/home/yuqifang/projects/MICCAI2022_summary/multarget_code/MFH/s2_joint_train_tea1_tea2_stu/checkpoints_pretrain/'
        # pretrain_dir = '/shenlab/lab_stor4/yuqifang/projects/MICCAI2022/multarget_code/mean_teacher_addpretrain/train_tea1_tea2_stu/server_0_40/checkpoints_pretrain_0_40/'
        ddcnet_stu = Model_stu(in_channels=1, num_class=1, edge_importance_weighting=True).to(device)
        ddcnet_stu.load_state_dict(torch.load(pretrain_dir + 'stu_fold_' + str(fold) + '_epoch_30.pth', map_location='cuda:0'))
        ddcnet_tea1 = Model_tea1(in_channels=1, num_class=1, edge_importance_weighting=True).to(device)
        ddcnet_tea1.load_state_dict(torch.load(pretrain_dir + 'tea1_fold_' + str(fold) + '_epoch_30.pth', map_location='cuda:0'))
        ddcnet_tea2 = Model_tea2(in_channels=1, num_class=1, edge_importance_weighting=True).to(device)
        ddcnet_tea2.load_state_dict(torch.load(pretrain_dir + 'tea2_fold_' + str(fold) + '_epoch_30.pth', map_location='cuda:0'))

        with open(SAVE_PATH + 'fold_' + str(fold) + '_train_loss_and_acc.txt', 'a') as f:
            f.write('total_epoch: {}, batch_size: {}, initial_lr {:.8}, drop_lr: {:.5}, drop_lr_per_epoch: {}\n'.format(TRAIN_EPOCHS, BATCH_SIZE, learning_rate, drop, epochs_drop))
        with open(SAVE_PATH + 'src_fold_' + str(fold) + '_test_loss_and_acc.txt', 'a') as f:
            f.write('total_epoch: {}, batch_size: {}, initial_lr {:.8}, drop_lr: {:.5}, drop_lr_per_epoch: {}\n'.format(TRAIN_EPOCHS, BATCH_SIZE, learning_rate, drop, epochs_drop))
        with open(SAVE_PATH + 'tgt1_fold_' + str(fold) + '_test_loss_and_acc.txt', 'a') as f:
            f.write('total_epoch: {}, batch_size: {}, initial_lr {:.8}, drop_lr: {:.5}, drop_lr_per_epoch: {}\n'.format(TRAIN_EPOCHS, BATCH_SIZE, learning_rate, drop, epochs_drop))
        with open(SAVE_PATH + 'tgt2_fold_' + str(fold) + '_test_loss_and_acc.txt', 'a') as f:
            f.write('total_epoch: {}, batch_size: {}, initial_lr {:.8}, drop_lr: {:.5}, drop_lr_per_epoch: {}\n'.format(TRAIN_EPOCHS, BATCH_SIZE, learning_rate, drop, epochs_drop))

        for epoch in range(1, TRAIN_EPOCHS + 1):
            print(f'Train Epoch {epoch}:')
            train_ddcnet_union(epoch, ddcnet_stu, ddcnet_tea1, ddcnet_tea2, src_loader, tgt1_train_loader, tgt2_train_loader, learning_rate, drop, epochs_drop)
            flag = 'src'
            correct_stu = test_ddcnet(ddcnet_stu, src_test_loader)
            flag = 'tgt1'
            correct_tgt1 = test_ddcnet(ddcnet_stu, tgt1_test_loader)
            flag = 'tgt2'
            correct_tgt2 = test_ddcnet(ddcnet_stu, tgt2_test_loader)

